!
!
! PING_PONG two-node message exchanges benchmark program
! 
! Contributed by Richard Frost <frost@SDSC.EDU>, caused problems
! on t3d with -device=t3d -arch=cray_t3d -no_short_longs -nodevdebug
! 
!
! This is a very time-consuming program on a workstation cluster.
! For this reason, I've modified it to do fewer tests (1/10 as many)
!
!
! This is a simple benchmark designed to measure the latency and bandwidth
! of a message-passing MIMD computer.  It is currently set up to run with
! MPI.
!
! Compile (MPI mpich version 1.0.11 or later) with
!     % mpif77 -o pong pong.f
!
! (mpif77 is a script that hides details about libraries from the user)
!
! Execute as
!     % mpirun -np 2 pong
!
! Make sure that ~mpi/bin is in your path.
!
! Note that the MPI-specific calls are:
!
!      FMPI_INIT
!      FMPI_COMM_RANK
!      FMPI_COMM_SIZE
!
!      FMPI_Wtime
!      FMPI_Wtick
!
!      FMPI_SEND
!      FMPI_RECV
!
!      FMPI_FINALIZE
!
! Some care needs to be taken in using the
! appropriate timing routine.  Check the value of FMPI_Wtick() to see if
! the clock resolution is reasonable for your tests.
!
! The benchmark measures
! the time to send a message of length N bytes from node 0 to node 1 and
! receive an acknowledging copy of that message which node 1 sends back to 
! node 0.  Note that node 1 waits for the completion of its receive before
! sending the message back to node 0. Note also that the program is not
! necessarily optimal any given system, but is intended
! to provide a reasonably transparent baseline measurement. 
! 
! For message lengths len (= num of doubles * sizedouble),
! a total of msgspersample ping-pong message exchanges are made,	
! and half of the average round-trip time (i.e. the one-way message
! time) is then fit by a linear function y(N) = a + b*N via a least squares
! linear regression.  The coefficient a is then interpreted as the latency
! (time to send a 0-length message) and b as the inverse bandwidth (i.e. 1/b =
! bandwidth in bytes/sec)
!
! The entire procedure is repeated twice, with the bandwidth, latency, and
! measured and fitted values of the message times reported for each instance.
!
! The underlying message passing performance characteristics of a 
! particular system may not necessarily be accurately modeled by the simple
! linear function assumed here.  This may be reflected in a dependency of
! the observed latency and bandwidth on the range of message sizes used.
!
! Original author:
! R. Leary, San Diego Supercomputer Center
! leary@sdsc.edu        9/20/94
!
! Modified for MPI     10/27/95
! frost@sdsc.edu
!
! Modified for f90 interface 18/3/97
! jim@orange.lowtem.hokudai.ac.jp
!
! =========================== program header =========================== 
!

program pong
  use mpi
  implicit none

  ! sizedouble = size in bytes of real element
  integer,  parameter :: sizedouble=8

  ! Note: set these parameters to one of 3 cases:
  !  1. size (each sample) < packetization length of architecture
  !  2. size (each sample) > packetization length of architecture
  !  3. size (1st sample) < packetization length of architecture
  !   & size (all others) > packetization length of architecture
  !
  !  Some known packetization lengths:
  !    Paragon            ~1500    bytes
  !    Cray T3D           ~1500    bytes 
  !    TCP/IP networks    256-1024 bytes
  !
  ! samples = the number of data points collected
  integer, parameter :: samples=40
  ! initsamplesize = # of elements transmitted in 1st sample
  integer, parameter :: initsamplesize=125
  ! samplesizeinc = sample size increase per iteration (linear rate)
  integer, parameter :: samplesizeinc=125
  !     parameter(samplesizeinc=1)
  ! msgspersample = the number of messages
  integer, parameter :: msgspersample=1000

  ! The buffer array contains the message , while x(i) is the message size 
  ! and y(i) the corresponding measured one-way average time. 
  ! Note that buffer is a real array
  !
  ! ibufcount = total number of elements in buffer
  integer, parameter :: ibufcount=(initsamplesize+((samples-1)*samplesizeinc))
  !
  real :: buffer(ibufcount)
  real :: x(samples), y(samples)
  real :: t1, t2
  real :: a, b, bandw
  real :: sumx, sumy, sumxx, sumxy
  real :: det, fit

  integer rank, numprocs
  integer idest
  integer i,sample
  integer num
  type(mpi_tag_t),parameter :: msg_tag = mpi_tag_t(0) ! tag for messages 
  !
  ! =========================== begin =========================== 
  !

  call FMPI_INIT()
  call FMPI_COMM_RANK(rank)
  call FMPI_COMM_SIZE(numprocs)

  if (numprocs /= 2) then
     print*,'This program is only valid for 2 processors'
     print*,'numprocs = ', numprocs
     call fmpi_abort
     stop 
  endif

  buffer =(/(i,i=1,ibufcount)/)

  if (fmpi_is_root(rank)) then
     print *, ' MPI pong test'
     print *, ' samples = ', samples
     print *, ' initsamplesize = ', initsamplesize
     print *, ' samplesizeinc = ', samplesizeinc
     print *, ' msgspersample = ', msgspersample
     print *, ' ibufcount = ', ibufcount
     print('(" clock resolution = ",e10.5)'), FMPI_Wtick() 
     print *
  endif

  call FMPI_BARRIER

  !
  ! =========================== main loop =========================== 
  !

  do sample = 1,samples
     num = initsamplesize + ((sample-1)*samplesizeinc)
     ! debug
     !print *, ' Rank=', rank,', sample =', sample,', num =', num
     call FMPI_BARRIER()
     ! Find initial elapsed time in seconds
     if(fmpi_is_root(rank)) then
        ! Send message from node 0 to 1 and receive message from 1
        idest = 1
        t1 = FMPI_Wtime()
        do i = 1,msgspersample
           call FMPI_SEND(buffer(1:num), idest, msg_tag)
           call FMPI_RECV(buffer(1:num), FMPI_ANY_SOURCE, FMPI_ANY_TAG)
        end do
        t2 = FMPI_Wtime()
     else
        ! Send message from node 1 to 0 and receive message from 0
        idest = 0
        t1 = FMPI_Wtime()
        do i = 1,msgspersample
           call FMPI_RECV(buffer(1:num), FMPI_ANY_SOURCE, FMPI_ANY_TAG)
           call FMPI_SEND(buffer(1:num), idest, msg_tag)
        end do
        t2 = FMPI_Wtime()
     endif

     ! independent variable is message length:
     x(sample) = num * sizedouble

     ! dependent variable is average one-way transit time:
     y(sample) = ((t2 - t1) * 0.5) / msgspersample

  end do

  ! now do linear least squares fit to data
  ! time = a + b*x

  if (fmpi_is_root(rank)) then
     sumy = sum(y)
     sumx = sum(x)
     sumxy = sum(x*y)
     sumxx = sum(x*x)
     det = (samples * sumxx) - (sumx * sumx)
     a = (1.d6 * ((sumxx * sumy) - (sumx * sumxy))) / det
     b = (1.d6 * ((samples * sumxy) - (sumx * sumy))) / det


     print *
     print *, ' least squares fit:  time = a + b * (msg length)'
     print("(5x,'a = latency = ',f8.2,' microseconds')"), a
     print("(5x,'b = inverse bandwidth = ' , f8.5,' secs/Mbyte')"), b
     bandw = 1./b
     print("(5x,'1/b = bandwidth = ',f8.2,' Mbytes/sec')"), bandw
     print *
     print *, '    message         observed          fitted'
     print *, ' length(bytes)     time(usec)       time(usec)'
     print *

     do i=1,samples
        fit = a + b*x(i) 
        y(i) = y(i)*1.d6
        print("(3x,f8.0,5x,f12.2,5x,f12.2)"), x(i),y(i),fit
     end do
  endif
call FMPI_FINALIZE
end program
